{{
  Bitmap 320x192 / Text 80x24 using Gameduino sprites
  A demo for Propeller ASC + Gameduino
  by Alessandro De Luca (AntoineDoinel)

  Based on "Bitmap" demo for Gameduino
  Copyright (c) 2011 by James Bowman <jamesb@excamera.com>
}}
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

OBJ
  bf : "Font_4x8"
'  fp : "Float32"
  fp : "F32"
  GD : "GD_ASM.spin"

CON
  RUN_TIME = 30

VAR
  long  seed, gcolor
  long  fontptr, CRLF, LFCR
  long  c_row, c_col, f_rev

PUB Main | i, n, d, t0, t1, ptr
  fp.start
  GD.begin
  GD.ascii
  screen_init

  fontptr := bf.get_font_ptr
  f_rev := 0
  c_col := 0
  c_row := 23
  LFCR  := 1
  CRLF  := 0
  gcolor := 4

  repeat
    restart(gcolor)
    test_loop
    waitcnt(cnt + clkfreq)
    out(11)
    test_text
    out(13)
    out(10)
    waitcnt(cnt + clkfreq)
    ptr := @ctext
    n := @ctext - @ctext_end
    repeat 1024
      out(byte[ptr++])
    waitcnt(cnt + clkfreq)
    out(12)
    repeat 1024
      out(byte[ptr++])
    waitcnt(cnt + clkfreq)
    gcolor := (gcolor + 1) // 6

DAT
ctext   file  "bitmap.c"
ctext_end

PRI replicate(color)
' replicate a 2-bit color across the whole byte.   Optimization for setpixel
  return (color << 6) | (color << 4) | (color << 2) | color

PRI setpixel(px, py, color) | addr, mask
' Set pixel at (x,y) to color.   (note that color is replicated).
'
'  Because of the way the sprites are laid out in setup(), it's not too
'  hard to translate the pixel (x,y) to an address and mask.  Taking the
'  two byte values as x7-x0 and y7-y0, the address of the pixel is:
'
'     x5 x4 y7 y6 y5 y4 y3 y2 y1 y0 x3 x2 x1 x0
'
' (x6, x7) gives the value of the mask.
'
  if (px < 0) or (px > 319) or (py < 0) or (py > 191)
    return
  addr := GD#RAM_SPRIMG + (((px >> 4) // 5) * 3072) + (py << 4) | (px & $0F)
  mask := $C0 >> ((px / 40) & 6)
  GD.m_wr(addr, (GD.m_rd(addr) & !mask) | (color & mask))

PRI SWAP(ptra, ptrb) | t
  t := long[ptra]
  long[ptra] := long[ptrb]
  long[ptrb] := t
{
PRI plot(x, y, color)
  color := replicate(color)
  setpixel(x, y, color)
}
PRI plot(x, y, color) | addr, mask
  color := (color << 6) | (color << 4) | (color << 2) | color
  addr := GD#RAM_SPRIMG + (((x >> 4) // 5) * 3072) + (y << 4) | (x & $0F)
  mask := $C0 >> ((x / 40) & 6)
  GD.m_wr(addr, (GD.m_rd(addr) & !mask) | (color & mask))

PRI line(x0, y0, x1, y1, color) | steep, deltax, deltay, error, ystep, tx, ty
' Draw color line from (x0,y0) to (x1,y1).
  color := replicate(color)
  deltax := ||(x1 - x0)
  deltay := ||(y1 - y0)
  if deltay > deltax
    steep := true
  else
    steep := false
  if steep
    SWAP(@x0, @y0)
    SWAP(@x1, @y1)
  if x0 > x1
    SWAP(@x0, @x1)
    SWAP(@y0, @y1)
  deltax :=    x1 - x0
  deltay := ||(y1 - y0)
  error  := deltax >> 1
  if y0 < y1    
    ystep :=  1
  else
    ystep := -1
  ty := y0
  x1 := x1 - 1
  repeat tx from x0 to x1
    if steep
      setpixel(ty, tx, color)
    else
      setpixel(tx, ty, color)
    error -= deltay
    if (error < 0)
      ty += ystep
      error += deltax

PRI random(limit) | rval
  rval := (?seed) & $7FFFFFFF
  return (rval // limit)

PRI RANDOM_RGB
'  return GD.RGB(cnt & 255, cnt & 255, cnt & 255)
  return GD.RGB(random(256), random(256), random(256))

PRI restart(pal) | i
  ' Clear the screen
  GD.fill(GD#RAM_PIC, 0, 4096)
  GD.putstr(11, 1, string("320x192 Bitmap Demonstration"))

  GD.putstr(8, 3, string("80x24 Monochrome Text and Graphics"))
  GD.putstr(16, 4, string("Using 4x8 AA Font"))

  ' Clear the screen
  GD.fill(GD#RAM_SPRIMG, 0, 16384)

  GD.putstr(13, 34, string("Palette:"))
  ' Set the palette
  case pal
    0:
      GD.putstr(23, 34, string("Yellow on Blue"))
      GD.m_wr16(GD#PALETTE4A + 0, GD.RGB($00, $00, $CC))
      GD.m_wr16(GD#PALETTE4A + 2, GD.RGB($55, $55, $88))
      GD.m_wr16(GD#PALETTE4A + 4, GD.RGB($AA, $AA, $44))
      GD.m_wr16(GD#PALETTE4A + 6, GD.RGB($FF, $FF, $00))
    1:
      GD.putstr(23, 34, string("Classic Green"))
      GD.m_wr16(GD#PALETTE4A + 0, GD.RGB($11, $33, $22))
      GD.m_wr16(GD#PALETTE4A + 2, GD.RGB($22, $77, $44))
      GD.m_wr16(GD#PALETTE4A + 4, GD.RGB($33, $BB, $66))
      GD.m_wr16(GD#PALETTE4A + 6, GD.RGB($44, $FF, $88))
    2:
      GD.putstr(23, 34, string("Classic Amber"))
      GD.m_wr16(GD#PALETTE4A + 0, GD.RGB($33, $22, $11))
      GD.m_wr16(GD#PALETTE4A + 2, GD.RGB($77, $44, $22))
      GD.m_wr16(GD#PALETTE4A + 4, GD.RGB($BB, $66, $33))
      GD.m_wr16(GD#PALETTE4A + 6, GD.RGB($FF, $88, $44))
    3:
      GD.putstr(23, 34, string("Leaky B & W"))
      GD.m_wr16(GD#PALETTE4A + 0, GD.RGB($11, $22, $33))
      GD.m_wr16(GD#PALETTE4A + 2, GD.RGB($55, $66, $77))
      GD.m_wr16(GD#PALETTE4A + 4, GD.RGB($99, $AA, $BB))
      GD.m_wr16(GD#PALETTE4A + 6, GD.RGB($DD, $EE, $FF))
    4:
      GD.putstr(23, 34, string("Grayscale"))
      GD.m_wr16(GD#PALETTE4A + 0, GD.RGB($00, $00, $00))
      GD.m_wr16(GD#PALETTE4A + 2, GD.RGB($55, $55, $55))
      GD.m_wr16(GD#PALETTE4A + 4, GD.RGB($AA, $AA, $AA))
      GD.m_wr16(GD#PALETTE4A + 6, GD.RGB($FF, $FF, $FF))
    other: '(4+)
      GD.putstr(23, 34, string("Reverse Video"))
      GD.m_wr16(GD#PALETTE4A + 0, GD.RGB($FF, $FF, $FF))
      GD.m_wr16(GD#PALETTE4A + 2, GD.RGB($AA, $AA, $AA))
      GD.m_wr16(GD#PALETTE4A + 4, GD.RGB($55, $55, $55))
      GD.m_wr16(GD#PALETTE4A + 6, GD.RGB($00, $00, $00))

PRI draw_frame(col) | ii, jj
  line(  0,   0, 319,   0, col)
  line(319,   0, 319, 191, col)
  line(319, 191,   0, 191, col)
  line(  0, 191,   0,   0, col)
  repeat ii from 1 to 21
    jj := ii * 16 
    line(jj, 1, jj, 190, col ^ %11)
  repeat ii from 1 to 15
    jj := ii * 16 
    line(1, jj, 318, jj, col ^ %11)
  line(  0,  95, 319,  95, col)
  line(319,  96,   0,  96, col)
  line(159,   0, 159, 191, col)
  line(160, 191, 160,   0, col)

PRI draw_box(x1, y1, x2, y2, col) | ii
  repeat ii from y1 to y2
    line(x1, ii, x2, ii, col)

PUB screen_init | i, sx, sy, img, pal
  ' Draw 256 sprites left to right, top to bottom, all in 4-color
  ' palette mode.   By doing them in column-wise order, the address
  ' calculation in setpixel is made simpler.
  ' First 64 use bits 0-1, next 64 use bits 2-4, etc.
  ' This gives a 256 x 256 4-color bitmap.
  seed := cnt
  repeat i from 0  to 239
    sx := 40 + 16 * (i / 12)
    sy := 54 + 16 * (i // 12)
    img := i // 60              ' image 0-63 */
    pal := 3 - (i / 60)         ' palettes bits in columns 3,2,1,0
    GD.sprite(i, sx, sy, img, $08 | (pal << 1), 0, 0)
  restart(0)
  gcolor := 0

PUB test_loop | x1, x2, y1, y2, col, xs, ys
' restart(gcolor)
  draw_frame(2)
  col := 0
  repeat 3                     
    xs := fp.FDiv(1.0, fp.FFloat(16 + random(64)))
    ys := fp.FFloat(16 + random(64))
    x1 := 0
    y1 := 96
    col++
    repeat x2 from 0 to 319
      y2 := 96 + fp.FRound(fp.FMul(fp.Sin(fp.FMul(fp.FFloat(x2), xs)), ys))
      line(x1, y1, x2, y2, col)
      x1 := x2
      y1 := y2
  waitcnt(cnt + clkfreq)
{
  GD.putstr(18, 3, string("Random Lines"))
  repeat 64
    x1 := 1 + random(318)
    y1 := 1 + random(190)
    x2 := 1 + random(318)
    y2 := 1 + random(190)
    col := random(4) 
    line(x1, y1, x2, y2, col)
  waitcnt(cnt + clkfreq)

' restart(gcolor)
  draw_frame(2)
  GD.putstr(18, 3, string("Random Boxes"))
  repeat 16
    x1 := 1 + random(254)
    y1 := 1 + random(142)
    x2 := x1 + random(64)
    y2 := y1 + random(48)
    col := random(4) 
    draw_box(x1, y1, x2, y2, col)
  waitcnt(cnt + clkfreq)
}
PUB test_text
  str(string(10, 9, "THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG...#&@%"))
  waitcnt(cnt + clkfreq)
  repeat 4
    out(8)
    waitcnt(cnt + (clkfreq >> 2))
  out(10)
  str(string(32, 9, "...just to get shot a few leaps later!", 10))
  waitcnt(cnt + clkfreq)
  str(string(9, 9, "Normal  : ", $20, $41, $42, $43, $20, 10))
  str(string(9, 9, "Reverse : ", $A0, $C1, $C2, $C3, $A0, 10))

{
PRI gdcopy(dst, src, len)
  len >>= 1
  repeat len
    GD.m_wr16(dst, GD.m_rd16(src))
    dst += 2
    src += 2
}
PRI drawchar(cn, rn, chr, rvf) | i, data, src, cx, cy
  if (cn < 0) or (cn > 79) or (rn < 0) or (rn > 23)
    return
  src := fontptr + (chr << 3)
  cx := cn << 2
  cy := rn << 3
  repeat i from 0 to 7
    data := byte[src][i]
    if rvf
      data ^= $FF
    plot(cx++, cy,  data >> 6     )
    plot(cx++, cy, (data >> 4) & 3)
    plot(cx++, cy, (data >> 2) & 3)
    plot(cx, cy++,  data       & 3)
    cx -= 3

CON
  LOAD_OFFS = (GD#COLLISION - 128)

DAT
tmpbuf   byte 0[128]
scr_offs long GD#RAM_SPRIMG, GD#RAM_SPRIMG+3072, GD#RAM_SPRIMG+6144, GD#RAM_SPRIMG+9216, GD#RAM_SPRIMG+12288

PRI scroll | i, ci, ri, dst, ofs
  repeat ri from 0 to 22
    ofs := (ri << 7) 
    repeat ci from 0 to 4
      dst := long[@scr_offs][ci] + ofs  
      GD.load_coll(dst - LOAD_OFFS, @tmpbuf, 128)
      GD.copy(dst, @tmpbuf, 128)
  repeat ci from 0 to 4
    dst := long[@scr_offs][ci]  
    GD.Fill(dst + 2944, 0, 128)

PRI home
  c_row := 0
  c_col := 0

PRI clrscr
  GD.Fill(GD#RAM_SPRIMG, 0, 16384)
  home

PRI newline
  if ++c_row == 24
    scroll
    c_row := 23

PRI out(chr)
  case chr
     8: if c_col
          c_col--
          out(32)
          c_col--
     9: repeat
          c_col++
        until ((c_col & 7) == 0)
        if (c_col == 80)
          out(10)
    10: newline
        if LFCR
          c_col := 0
    11: home
    12: clrscr
    13: c_col := 0
        if CRLF
          newline
    other:
        drawchar(c_col, c_row, chr & 127, chr >> 7)
        if ++c_col == 80
          c_col := 0
          newline

PRI str(ptr) | i
  repeat i from 0 to strsize(ptr) - 1
    out(byte[ptr][i])